#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim.atoms import Atom
from udkm1Dsim.unitCell import UnitCell
from udkm1Dsim.structure import Structure
from pint import UnitRegistry
u = UnitRegistry()
u.default_format = '~P'


def test_structure():
    Dy = Atom('Dy')
    uc = UnitCell('uc', 'Unit Cell', 3.1*u.angstrom, heat_capacity=10*(u.J/u.kg/u.K),
                  lin_therm_exp=1e-6/u.K, therm_cond=1*(u.W/u.m/u.K),
                  opt_pen_depth=11*u.nm, sound_vel=5*(u.nm/u.ps))
    uc.add_atom(Dy, 'lambda strain: 0*(strain+1)')
    uc.add_atom(Dy, 'lambda strain: 0.5*(strain+1)')

    S = Structure('sample')
    assert S.name == 'sample'



from pint import UnitRegistry
u = UnitRegistry()
u.default_format = '~P'
Q_ = u.Quantity
from .atoms import Atom, AtomMixed
from .unitCell import UnitCell
from .structure import Structure

__all__ = ['Atom', 'AtomMixed', 'UnitCell', 'Structure', 'u']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the udkm1Dsimpy module.
#
# udkm1Dsimpy is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2017 Daniel Schick

"""A :mod:`Structure` module """

__all__ = ["Structure"]

__docformat__ = "restructuredtext"

import itertools
import numpy as np
from .unitCell import UnitCell


class Structure:
    """Structure

    The structure class can hold various sub_structures. Each
    sub_structure can be either a layer of N unitCell objects or a
    structure by itself.
    Thus it is possible to recursively build up 1D structures.

    Args:
        name (str): name of the sample

    Attributes:
        name (str): name of sample
        sub_structures (list): list of structures in sample
        substrate (object): structure of the substrate
        num_sub_systems (int): number of subsystems for heat and phonons
           (electronic, lattice, spins, ...)

    """

    def __init__(self, name):
        self.name = name
        self.num_sub_systems = 1
        self.sub_structures = []
        self.substrate = []

    def __str__(self):
        """String representation of this class"""
        return "hello world"
#        % check for the numbers of tabs added at the beginning
#            if nargin < 2
#                tabs = 0;
#            end%if
#            tabstr = '';
#            for i = 1:tabs
#                tabstr = strcat(tabstr, '\t');
#            end%for
#            % plot the properties
#            fprintf([tabstr 'Structure properties:\n']);
#            fprintf([tabstr 'Name  : %s\n'], obj.name);
#            fprintf([tabstr 'Length: %0.2f nm\n'], obj.getLength()/units.nm);
#            fprintf([tabstr '----\n']);
#            % traverse all substructures
#            for i = 1:size(obj.substructures,1)
#                if isa(obj.substructures{i,1},'unitCell')
#                    % the substructure is an unitCell
#                    fprintf([tabstr '%d times %s: %0.2f nm\n'], obj.substructures{i,2},...
#                        obj.substructures{i,1}.name,
#                        obj.substructures{i,2}*obj.substructures{i,1}.cAxis/units.nm);
#                else
#                    % the substructure is a structure instance by itself
#                    % call the display() method recursively
#                    fprintf([tabstr 'SubStructure %d times:\n'],...
#                        obj.substructures{i,2});
#                    obj.substructures{i,1}.displayTab(tabs+1);
#                end%if
#            end%for
#            fprintf([tabstr '----\n']);
#            % check for a substrate
#            if isa(obj.substrate,'structure')
#                fprintf([tabstr 'Substrate:\n']);
#                fprintf([tabstr '----\n']);
#                fprintf([tabstr '%d times %s: %0.2f nm\n'], obj.substrate.substructures{1,2},...
#                    obj.substrate.substructures{1,1}.name, ...
#                    obj.substrate.substructures{1,2}*obj.substrate.substructures{1,1}.cAxis/units.nm);
#            end%if
#        output = [['id', self.id],
#                  ['name', self.name],
#                  ['a-axis', '{:.4~P}'.format(self.a_axis)],
#                  ['b-axis', '{:.4~P}'.format(self.b_axis)],
#                  ['c-axis', '{:.4~P}'.format(self.c_axis)],
#                  ['area', '{:.4~P}'.format(self.area.to('angstrom**2'))],
#                  ['volume', '{:.4~P}'.format(self.volume.to('angstrom**3'))],
#                  ['mass', '{:.4~P}'.format(self.mass)],
#                  ['density', '{:.4~P}'.format(self.density.to('kg/meter**3'))],
#                  ['Debye Waller Factor', self.deb_wal_fac.to('meter**2')],
#                  ['sound velocity', '{:.4~P}'.format(self.sound_vel.to('meter/s'))],
#                  ['spring constant', self.spring_const * u.kg/u.s**2],
#                  ['phonon damping', self.phonon_damping.to('kg/s')],
#                  ['opt. pen. depth', self.opt_pen_depth.to('nm')],
#                  ['opt. refractive index', self.opt_ref_index],
#                  ['opt. ref. index/strain', self.opt_ref_index_per_strain],
#                  ['thermal conduct.', ' W/(m K)\n'.join(self.therm_cond_str) + ' W/(m K)'],
#                  ['linear thermal expansion', '\n'.join(self.lin_therm_exp_str)],
#                  ['heat capacity', ' J/(kg K)\n'.join(self.heat_capacity_str) + ' J/(kg K)'],
#                  ['subsystem coupling', ' W/m³\n'.join(self.sub_system_coupling_str) + ' W/m³']]
#
#        class_str = 'Unit Cell with the following properties\n\n'
#        class_str += tabulate(output, headers=['parameter', 'value'], tablefmt="rst",
#                              colalign=('right',), floatfmt=('.2f', '.2f'))
#        class_str += '\n\n' + str(self.num_atoms) + ' Constituents:\n'
#
#        atoms_str = []
#        for i in range(self.num_atoms):
#            atoms_str.append([self.atoms[i][0].name,
#                              '{:0.2f}'.format(self.atoms[i][1](0)),
#                              self.atoms[i][2]])
#        class_str += tabulate(atoms_str, headers=['atom', 'position', 'position function'],
#                              tablefmt="rst")

    def visualize(self):
        """visualize"""
#        % initialize input parser and define defaults and validators
#            p = inputParser;
#            p.addRequired('obj'      , @(x)isa(x,'structure'));
#            p.addParamValue('handle', '', @ishandle);
#            % parse the input
#            p.parse(obj,varargin{:});
#            % assign parser results to object properties
#            if isempty(p.Results.handle)
#                h = figure;
#            else
#                h = p.Results.handle;
#            end%if
#            a = obj.getUniqueUnitCells();
#            N = size(a,1);
#            figure(h);
#            distances = obj.getDistancesOfUnitCells/units.nm;
#            stairs(distances,obj.getUnitCellVectors, 'LineWidth', 2);
#            axis([min(distances) max(distances) 0.9 length(obj.getUniqueUnitCells)+0.1]);
#            xlabel('Distance [nm]');
#            title('Structure Visualization');
#            set(gca,'YTick',1:N,'YTickLabel', a(:,1));
        pass

    def get_hash(self):
        """hash"""
#        UCs = obj.getUniqueUnitCells();
#            param = cell(size(UCs,1),1);
#            for i=1:size(UCs,1)
#                param{i} = UCs{i,2}.getPropertyStruct(varargin{:});
#            end%for
#            [~, IDs] = obj.getUnitCellVectors();
#            param(end+1) = {IDs};
#            % dataHash is an external function
#            hash = dataHash(param);
        pass

    def add_sub_structure(self, sub_structure, N):
        """add_sub_structure

        Add a sub_structure of N unitCells or N structures to the
        structure.

        Args:
            sub_structure (UnitCell, Structure): unit cell or structure
               to add as sub structure
            N (int): number or repetitions

        """
        # check of the sub_structure is an instance of the unitCell of
        # structure class
        if not isinstance(sub_structure, (UnitCell, Structure)):
            raise ValueError('Class '
                             + type(sub_structure).__name__
                             + ' is no possible sub structure.'
                             + 'Only UnitCell and'
                             + 'Structure classes are allowed!')

        # if a structure is added as a sub_structure, the sub_structure
        # can not have a substrate
        if isinstance(sub_structure, Structure):
            if sub_structure.substrate:
                raise ValueError('No substrate in sub_structure allowed!')

        # check the number of subsystems of the sub_structure
        if ((self.num_sub_systems > 1)
           and not (sub_structure.num_sub_systems == self.num_sub_systems)):

            raise ValueError('The number of subsystems in each sub_structure'
                             'must be the same!')
        else:
            self.num_sub_systems = sub_structure.num_sub_systems

        # add a sub_structure of N repetitions to the structure with
        self.sub_structures.append([sub_structure, N])

    def add_substrate(self, sub_structure):
        """add_substrate

        Add a structure as static substrate to the structure

        Args:
            sub_structure (Structure): substrate structure

        """
        if not isinstance(sub_structure, Structure):
            raise ValueError('Class '
                             + type(sub_structure).__name__
                             + ' is no possible substrate. '
                             + 'Only structure class is allowed!')

        self.substrate = sub_structure

    def get_number_of_sub_structures(self):
        """get_number_of_sub_structures

        Returns the number of all sub structures.
        This methods does not return the number of all unitCells in the
        structure, see get_number_of_unit_cells().

        """
        N = 0
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], UnitCell):
                N = N + 1
            else:
                N = N + self.sub_structures[i][0].getNumberOfsub_structures()
        return N

    def get_number_of_unit_cells(self):
        """get_number_of_unit_cells

        Returns the number of all unitCells in the structure.

        """
        N = 0
        # traverse the substructres
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], UnitCell):
                N = N + self.sub_structures[i][1]
            else:
                # its a sturcture, so call the method recursively
                N = N + self.sub_structures[i][0].getNumberOfUnitCells()*self.sub_structures[i][1]

        return N

    def get_number_of_unique_unit_cells(self):
        """get_number_of_unique_unit_cells

        Returns the number of unique unitCells in the structure.

        """
        N = len(self.get_unique_unit_cells()[0])
        return N

    def get_length(self):
        """get_length

        Returns the length from surface to bottom of the structure

        """
        pass

    def get_unique_unit_cells(self):
        """get_unique_unit_cells

        Returns a list of ids and handles of all unique UnitCell
        instances in the structure.
        The uniqueness is determined by the handle of each unitCell
        instance.

        """
        uc_ids = []
        uc_handles = []
        # traverse the sub_structures
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], UnitCell):
                # its a UnitCell
                uc_id = self.sub_structures[i][0].id
                if not uc_ids:
                    # the cell array is empty at the beginning so add
                    # the first unitCell
                    uc_ids = uc_ids + [uc_id]
                    uc_handles = uc_handles + [self.sub_structures[i][0]]
                else:
                    # the cell array is not empty so check if the id is
                    # already in the ucs id vector
                    if uc_id not in uc_ids:
                        # if id not in list, so add it
                        uc_ids = uc_ids + [uc_id]
                        uc_handles = uc_handles + [self.sub_structures[i][0]]
            else:
                # its a sub_structure
                if not uc_ids:
                    # the cell array is empty at the beginning so call
                    # the method recursively and add the result to the
                    # ucs array
                    uc_ids = self.sub_structures[i][0].get_unique_unit_cells()[0]
                    uc_handles = self.sub_structures[i][0].get_unique_unit_cells()[1]
                else:
                    # the cell array is not empty so check if the ids
                    # from the recursive call are already in the ucs id
                    # vector.
                    temp1 = self.sub_structures[i][0].get_unique_unit_cells()[0]
                    temp2 = self.sub_structures[i][0].get_unique_unit_cells()[1]
                    for j, temp in enumerate(temp1):
                        # check all ids from recursive call
                        if temp1[j] not in uc_ids:
                            # ids not in list, so add them
                            uc_ids = uc_ids + [temp1[j]]
                            uc_handles = uc_handles + [temp2[j]]

        return uc_ids, uc_handles

    def get_unit_cell_vectors(self, *args):
        """get_unit_cell_vectors

        Returns three lists with the numeric index of all unit cells
        in a structure given by the get_unique_unit_cells() method and
        addidionally vectors with the ids and Handles of the
        corresponding unitCell instances.
        The list and order of the unique unitCells can be either handed
        as an input parameter or is requested at the beginning.

        Args:
            ucs (Optional[list]): list of unique unit cells including
               ids and handles

        """
        indices = []
        uc_ids = []
        uc_handles = []
        # if no ucs (UniqueUnitCells) are given, we have to get them
        if len(args) < 1:
            ucs = self.get_unique_unit_cells()
        else:
            ucs = args[0]
        # traverse the substructres
        for i in range(len(self.sub_structures)):
            if isinstance(self.sub_structures[i][0], UnitCell):
                # its a UnitCell
                # find the index of the current UC id in the unique
                # unitCell vector
                Index = ucs[0].index(self.sub_structures[i][0].id)
                # add the index N times to the indices vector
                indices = np.append(indices, Index*np.ones(self.sub_structures[i][1]))
                # create a cell array of N unitCell ids and add them to
                # the ids cell array
                temp1 = list(itertools.repeat(self.sub_structures[i][0].id,
                                              self.sub_structures[i][1]))
                uc_ids = uc_ids + list(temp1)
                # create a cell array of N unitCell handles and add them to
                # the Handles cell array
                temp2 = list(itertools.repeat(self.sub_structures[i][0],
                                              self.sub_structures[i][1]))
                uc_handles = uc_handles + list(temp2)
            else:
                # its a structure
                # make a recursive call and hand in the same unique
                # unit cell vector as we used before
                [temp1, temp2, temp3] = self.sub_structures[i][0].get_unit_cell_vectors(ucs)
                temp11 = []
                temp22 = []
                temp33 = []
                # concat the temporary arrays N times
                for j in range(self.sub_structures[i][1]):
                    temp11 = temp11 + list(temp1)
                    temp22 = temp22 + list(temp2)
                    temp33 = temp33 + list(temp3)
                # add the temporary arrays to the outputs
                indices = np.append(indices, temp11)
                uc_ids = uc_ids + list(temp22)
                uc_handles = uc_handles + list(temp33)
        return indices, uc_ids, uc_handles

    def get_all_positions_per_unique_unit_cell(self):
        """get_all_positions_per_unique_unit_cell

        Returns a list with one vector of position indices for
        each unique unitCell in the structure.

        """
        ucs = self.get_unique_unit_cells()
        indices = self.get_unit_cell_vectors()[0]
        pos = {}  # Dictionary used instead of array
        for i, uc in enumerate(ucs[0]):
            pos[ucs[0][i]] = list(np.where(indices == i))
        # Each element accessible through Unit cell id
        return pos

    def get_distances_of_unit_cells(self):
        """get_distances_of_unit_cells

        Returns a vector of the distance from the surface for each unit
        cell starting at 0 (dStart) and starting at the end of the first
        unit cell (dEnd) and from the center of each unit cell (dMid).

        """
        c_axes = self.get_unit_cell_property_vector('_c_axis')
        d_end = np.cumsum(c_axes)
        d_start = np.hstack([[0], d_end[0:-1]])
        d_mid = (d_start + c_axes)/2
        return d_start, d_end, d_mid

        def get_distances_of_interfaces(self):
            """get_distances_of_interfaces"""
#            % Returns the distances from the surface of each interface of the
#            % structure.
#            function [distIntf Indices] = getDistancesOfInterfaces(obj)
#                [dStart dEnd]   = obj.getDistancesOfUnitCells();
#                Indices         = [1 diff(obj.getUnitCellVectors())'];
#                distIntf        = [dStart(Indices ~= 0)' dEnd(end)]';
#            end
            pass

        def interp_distance_at_interfaces(self):
            """interp_distance_at_interfaces"""
#            % Returns a distance Vector of the center of UCs interpolated by an
#            % odd number N at the interface of sturctures.
#            function [distInterp originalIndicies] = interpDistanceAtInterfaces(obj,N)
#                [dStart,dEnd,dMid] = obj.getDistancesOfUnitCells();
#                % these are the distances of the interfaces
#                distIntf = obj.getDistancesOfInterfaces();
#                % we start with the distances of the centers of the unit cells
#                distInterp = dMid;
#
#                N = floor(N); % make N an integer
#                if mod(N,2) == 0
#                    % we want to have odd numbers
#                    N = N+1;
#                end%if
#
#                % traverse all distances
#                for i=1:length(distIntf)
#                    x = distIntf(i); % this is the distance of an interface
#
#                    inda = finderb(x,dStart); % this is the index of an UC after the interface
#                    indb = inda-1; % this is the index of an UC before the interface
#
#                    % now interpolate linearly N new distances at the interface
#                    if indb == 0 % this is the surface interface
#                        distInterp = vertcat(distInterp,linspace(0,dMid(inda),2+(N-1)/2)');
#                    elseif inda >= length(dMid) % this is the bottom interface
#                        distInterp = vertcat(distInterp,
#                       linspace(dMid(inda),dEnd(end),2+(N-1)/2)');
#                    else % this is a surface inside the structure
#                        distInterp = vertcat(distInterp,linspace(dMid(indb),dMid(inda),2+N)');
#                    end%if
#                end%for
#
#                distInterp = unique(sort(distInterp)); % sort and unify the distances
#                % these are the indicies of the original distances in the interpolated new vector
#                originalIndicies = finderb(dMid,distInterp);
#            end%function
            pass

    def get_unit_cell_property_vector(self, property_name):
        """get_unit_cell_property_vector

        Returns a vector for a property of all unitCells in the
        structure. The property is determined by the propertyName and
        returns a scalar value or a function handle.

        Args:
            property_name (str): type of property to return as vector

        """
        # get the Handle to all unitCells in the Structure
        handles = self.get_unit_cell_vectors()[2]

        if callable(getattr(handles[0], property_name)):
            # it's a function
            print("it's a function")
            prop = np.zeros([self.get_number_of_unit_cells()])
            for i in range(self.get_number_of_unit_cells()):
                prop[i] = getattr(handles[i], property_name)
        elif ((type(getattr(handles[0], property_name)) is list) or
                (type(getattr(handles[0], property_name)) is str)):
            # it's a list if functions or str
            prop = []
            for i in range(self.get_number_of_unit_cells()):
                # Prop = Prop + getattr(Handles[i],types)
                prop.append(getattr(handles[i], property_name))
        else:
            # its a number or array
            ucs = self.get_unique_unit_cells()
            temp = np.zeros([len(ucs[0]), 1])
            for i, uc in enumerate(ucs[1]):
                temp[i] = len(getattr(uc, property_name))
            prop = np.zeros([self.get_number_of_unit_cells(), int(np.max(temp))])
            del temp
            # traverse all unitCells
            for i in range(self.get_number_of_unit_cells()):
                temp = getattr(handles[i], property_name)
                prop[i, 0:len(temp)] = temp

        return prop

    def get_unit_cell_handle(self, i):
        """get_unit_cell_handle

        Returns the handle to the unitCell at position i in the
        structure.

        """
        handles = self.get_unit_cell_vectors()[2]
        handle = handles[i]
        return handle



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim.atoms import Atom
from udkm1Dsim.unitCell import UnitCell
from pint import UnitRegistry
u = UnitRegistry()
u.default_format = '~P'


def test_unit_cell():
    Dy = Atom('Dy')
    uc = UnitCell('uc', 'Unit Cell', 3.1*u.angstrom, heat_capacity=10*(u.J/u.kg/u.K),
                  lin_therm_exp=1e-6/u.K, therm_cond=1*(u.W/u.m/u.K),
                  opt_pen_depth=11*u.nm, sound_vel=5*(u.nm/u.ps))
    uc.add_atom(Dy, 'lambda strain: 0*(strain+1)')
    uc.add_atom(Dy, 'lambda strain: 0.5*(strain+1)')

    assert uc.id == 'uc'
    assert uc.name == 'Unit Cell'
    assert uc.a_axis == 3.1*u.angstrom
    assert uc.b_axis == 3.1*u.angstrom
    assert uc.c_axis == 3.1*u.angstrom
    assert uc.heat_capacity[0](300) == 10
    assert uc.int_heat_capacity[0](300) == 3000
    assert uc.lin_therm_exp[0](300) == 1e-6
    assert uc.int_lin_therm_exp[0](300) == 0.0003
    assert uc.therm_cond[0](300) == 1
    assert uc.opt_pen_depth == 11*u.nm
    assert uc.sound_vel == 5*(u.nm/u.ps)
    assert uc.get_property_dict(types='phonon') == {'_c_axis': 3.1e-10,
                                                    '_mass': 5.615766784599375e-26,
                                                    '_phonon_damping': 0.0,
                                                    'int_lin_therm_exp_str':
                                                        ['lambda T : 1.0e-6*T'],
                                                    'num_sub_systems': 1}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the udkm1Dsimpy module.
#
# udkm1Dsimpy is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2019 Daniel Schick

__all__ = ["UnitCell"]

__docformat__ = "restructuredtext"

import numpy as np
from inspect import isfunction
from sympy import integrate, Symbol
from sympy.utilities.lambdify import lambdify
from tabulate import tabulate
from . import u, Q_


class UnitCell:
    """UnitCell

    The unitCell class hold different structural properties of real
    physical unit cells and also an array of atoms at different postions
    in the unit cell.

    Args:
        id (str): id of the UnitCell
        name (str): name of the UnitCell
        c_axis (float): c-axis of the UnitCell

    Keyword Args:
        a_axis (float): a-axis of the UnitCell
        b_axis (float): b-axis of the UnitCell
        deb_wal_fac (float): Debye Waller factor
        sound_vel (float): sound velocity
        phonon_damping (float): phonon damping
        opt_pen_depth (float): optical penetration depth
        opt_ref_index (float): refractive index
        opt_ref_index_per_strain (float): change of refractive index per
           strain
        heat_capacity (float): heat capacity
        therm_cond (float): thermal conductivity
        lin_therm_exp (float): linear thermal expansion
        sub_system_coupling (float): sub-system coupling

    Attributes:
        id (str): id of the unit cell
        name (str): name of the unit cell
        atoms (list[atom, @lambda]): list of atoms and funtion handle
           for strain dependent displacement
        num_atoms (int): number of atoms in unit cell
        deb_wal_fac (float): Debye Waller factor <u>² [m²]
        spring_const (ndarray[float]): spring constant of the unit cell
           [kg/s²] and higher orders
        opt_ref_index (ndarray[float]): optical refractive index - real
           and imagenary part :math:`n + i\kappa`
        opt_ref_index_per_strain (ndarray[float]): optical refractive
           index change per strain - real and imagenary part
           :math:`\\frac{d n}{d \eta} + i\\frac{d \kappa}{d \eta}`
        therm_cond (list[@lambda]): list of HANDLES T-dependent thermal
           conductivity [W/(m K)]
        lin_therm_exp (list[@lambda]): list of HANDLES T-dependent
           linear thermal expansion coefficient (relative)
        int_lin_therm_exp (list[@lambda]): list of HANDLES T-dependent
           integrated linear thermal expansion coefficient
        heat_capacity (list[@lambda]): list of HANDLES T-dependent heat
           capacity function [J/(kg K)]
        int_heat_capacity (list[@lambda]): list of HANDLES T-dependent
           integrated heat capacity function
        sub_system_coupling (list[@lambda]): list of HANDLES of coupling
           functions of different subsystems [W/m³]
        num_sub_systems (int): number of subsystems for heat and phonons
           (electrons, lattice, spins, ...)

    """

    def __init__(self, id, name, c_axis, **kwargs):
        self.id = id
        self.name = name
        self.c_axis = c_axis
        self.a_axis = kwargs.get('a_axis', self.c_axis)
        self.b_axis = kwargs.get('b_axis', self.a_axis)
        self.atoms = []
        self.num_atoms = 0
        self.mass = 0*u.kg
        self.density = 0*u.kg/u.m**2
        self.spring_const = np.array([0])
        self.deb_wal_fac = kwargs.get('deb_wal_fac', 0*u.m**2)
        self.sound_vel = kwargs.get('sound_vel', 0)
        self.phonon_damping = kwargs.get('phonon_damping', 0*u.kg/u.s)
        self.opt_pen_depth = kwargs.get('opt_pen_depth', 0*u.nm)
        self.opt_ref_index = kwargs.get('opt_ref_index', 0)
        self.opt_ref_index_per_strain = kwargs.get('opt_ref_index_per_strain', 0)
        self.heat_capacity, self.heat_capacity_str = self.check_cell_array_input(
                kwargs.get('heat_capacity', 0))
        self.therm_cond, self.therm_cond_str = self.check_cell_array_input(
                kwargs.get('therm_cond', 0))
        self.lin_therm_exp, self.lin_therm_exp_str = self.check_cell_array_input(
                kwargs.get('lin_therm_exp', 0))
        self.sub_system_coupling, self.sub_system_coupling_str = self.check_cell_array_input(
                kwargs.get('sub_system_coupling', 0))

        if (len(self.heat_capacity) == len(self.therm_cond)
            and len(self.heat_capacity) == len(self.lin_therm_exp)
                and len(self.heat_capacity) == len(self.sub_system_coupling)):
            self.num_sub_systems = len(self.heat_capacity)
        else:
            raise ValueError('Heat capacity, thermal conductivity, linear'
                             'thermal expansion and subsystem coupling have not'
                             'the same number of elements!')

        self.area = self.a_axis * self.b_axis
        self.volume = self.area * self.c_axis

    def __str__(self):
        """String representation of this class"""
        output = [['id', self.id],
                  ['name', self.name],
                  ['a-axis', '{:.4~P}'.format(self.a_axis)],
                  ['b-axis', '{:.4~P}'.format(self.b_axis)],
                  ['c-axis', '{:.4~P}'.format(self.c_axis)],
                  ['area', '{:.4~P}'.format(self.area.to('angstrom**2'))],
                  ['volume', '{:.4~P}'.format(self.volume.to('angstrom**3'))],
                  ['mass', '{:.4~P}'.format(self.mass)],
                  ['density', '{:.4~P}'.format(self.density.to('kg/meter**3'))],
                  ['Debye Waller Factor', self.deb_wal_fac.to('meter**2')],
                  ['sound velocity', '{:.4~P}'.format(self.sound_vel.to('meter/s'))],
                  ['spring constant', self.spring_const * u.kg/u.s**2],
                  ['phonon damping', self.phonon_damping.to('kg/s')],
                  ['opt. pen. depth', self.opt_pen_depth.to('nm')],
                  ['opt. refractive index', self.opt_ref_index],
                  ['opt. ref. index/strain', self.opt_ref_index_per_strain],
                  ['thermal conduct.', ' W/(m K)\n'.join(self.therm_cond_str) + ' W/(m K)'],
                  ['linear thermal expansion', '\n'.join(self.lin_therm_exp_str)],
                  ['heat capacity', ' J/(kg K)\n'.join(self.heat_capacity_str) + ' J/(kg K)'],
                  ['subsystem coupling', ' W/m³\n'.join(self.sub_system_coupling_str) + ' W/m³']]

        class_str = 'Unit Cell with the following properties\n\n'
        class_str += tabulate(output, headers=['parameter', 'value'], tablefmt="rst",
                              colalign=('right',), floatfmt=('.2f', '.2f'))
        class_str += '\n\n' + str(self.num_atoms) + ' Constituents:\n'

        atoms_str = []
        for i in range(self.num_atoms):
            atoms_str.append([self.atoms[i][0].name,
                              '{:0.2f}'.format(self.atoms[i][1](0)),
                              self.atoms[i][2]])
        class_str += tabulate(atoms_str, headers=['atom', 'position', 'position function'],
                              tablefmt="rst")
        return class_str

    def clone_2_multiple(self, N, *args):
        """clone_2_multiple"""
#        %% clone2multiple
#        % Returns a cloned unit cell N time repeated. Accordingly, all
#        % physical properties are adapted, *despite of any specific function
#        % handle of the atomic position in the unit cell and higher orders
#        % of spring constants!*
#        % If no _ID_ or _name_ is given for the cloned unit cell, the name
#        % of the current unit cell is taken plus $N$.
#        function clone = clone2multiple(obj,N,varargin)
#            % initialize input parser and define defaults and validators
#            p = inputParser;
#            p.addRequired('N'                                       , @isnumeric);
#            p.addParamValue('ID'            , [obj.ID num2str(N)]   , @ischar);
#            p.addParamValue('name'          , [obj.name num2str(N)] , @ischar);
#            % parse the input
#            p.parse(N,varargin{:});
#            % assign parser results
#            N       = p.Results.N;
#            newID   = p.Results.ID;
#            newName = p.Results.name;
#            % initialize the cloned object with required inputs
#            clone = unitCell(obj.ID,obj.name,obj.cAxis);
#            % copy all properties from the current to the cloned object
#            props = properties(obj);
#            for i = 1:length(props)
#                clone.(props{i}) = obj.(props{i});
#            end%for
#            % reset some _N_-dependent properties
#            clone.ID = newID;
#            clone.name = newName;
#            clone.cAxis = N*obj.cAxis;
#            clone.volume = N*obj.volume;
#            % reset all atom-dependent properties
#            clone.atoms = {};
#            clone.mass = 0;
#            clone.density = 0;
#            clone.springConst = 0;
#            clone.numAtoms = 0;
#            % add all atoms to cloned unit cell, but the position handle is
#            % transfered!
#            for i = 0:N-1
#               for j = 1:obj.numAtoms
#                  clone.addAtom(obj.atoms{j,1},obj.atoms{j,2}(0)/N+i/N);
#               end%for
#            end%for
#        end%function
        pass

    def visualize(self, **kwargs):
        """visualize

        Allows for 3D presentation of unit cell by allow for a & b
        coordinate of atoms.
        Also add magnetization per atom.

        Todo: use the avogadro project as plugin
        Todo: create unit cell from CIF file e.g. by xrayutilities
        plugin.

        """
        import matplotlib.pyplot as plt
        import matplotlib.cm as cmx

        strains = kwargs.get('strains', 0)
        if not isinstance(strains, np.ndarray):
            strains = np.array([strains])

        colors = [cmx.Dark2(x) for x in np.linspace(0, 1, self.num_atoms)]
        atom_ids = self.get_atom_ids()

        for strain in strains:
            plt.figure()
            atoms_plotted = np.zeros_like(atom_ids)
            for j in range(self.num_atoms):
                if not atoms_plotted[atom_ids.index(self.atoms[j][0].id)]:
                    label = self.atoms[j][0].id
                    atoms_plotted[atom_ids.index(self.atoms[j][0].id)] = True
                    plt.plot(1+j, self.atoms[j][1](strain), 'o',
                             MarkerSize=10,
                             markeredgecolor=[0, 0, 0],
                             markerfaceColor=colors[atom_ids.index(self.atoms[j][0].id)],
                             label=label)
                else:
                    label = '_nolegend_'
                    plt.plot(1+j, self.atoms[j][1](strain), 'o',
                             MarkerSize=10,
                             markeredgecolor=[0, 0, 0],
                             markerfaceColor=colors[atom_ids.index(self.atoms[j][0].id)],
                             label=label)

            plt.axis([0.1, self.num_atoms+0.9, -0.1, (1.1+np.max(strains))])
            plt.grid(True)

            plt.title('Strain: {:0.2f}%'.format(strain))
            plt.ylabel('relative Position')
            plt.xlabel('# Atoms')
            plt.legend()
            plt.show()

    def get_property_dict(self, **kwargs):
        """get_property_dict

        Returns a dictionary with all parameters. objects or dicts and
        objects are converted to strings. if a type is given, only these
        properties are returned.

        """
        # initialize input parser and define defaults and validators
        types = ['all', 'heat', 'phonon', 'XRD', 'optical']
        properties_by_types = {'heat': ['_c_axis', '_area', '_volume', '_opt_pen_depth',
                                        'therm_cond_str', 'heat_capacity_str',
                                        'int_heat_capacity_str', 'sub_system_coupling_str',
                                        'num_sub_systems'],
                               'phonon': ['num_sub_systems', 'int_lin_therm_exp_str', '_c_axis',
                                          '_mass', '_spring_const', '_phonon_damping'],
                               'XRD': ['num_atoms', 'atoms', '_area', '_deb_wal_fac', '_c_axis'],
                               'optical': ['_c_axis', '_opt_pen_depth', 'opt_ref_index',
                                           'opt_ref_index_per_strain'],
                               }

        types = kwargs.get('types', 'all')
        attrs = vars(self)
        # define the property names by the given type
        if types == 'all':
            S = attrs
        else:
            S = dict((key, value) for key, value in attrs.items()
                     if key in properties_by_types[types])

        return S

    def check_cell_array_input(self, inputs):
        """ check_cell_array_input

        Checks the input for inputs which are cell arrays of function
        handles, such as the heat capacity which is a cell array of N
        function handles.

        """
        output = []
        outputStrs = []
        # if the input is not a list, we convert it to one
        if not isinstance(inputs, list):
            inputs = [inputs]
        # traverse each list element and convert it to a function handle
        for input in inputs:
            if isfunction(input):
                raise ValueError('Please use string representation of function!')
                output.append(input)
                outputStrs.append('no str representation available')
            elif isinstance(input, str):
                try:
                    output.append(eval(input))
                    outputStrs.append(input)
                except Exception as e:
                    print('String input for unit cell property ' + input + ' \
                        cannot be converted to function handle!')
                    print(e)
            elif isinstance(input, (int, float)):
                output.append(eval('lambda T: {:f}'.format(input)))
                outputStrs.append('lambda T: {:f}'.format(input))
            elif isinstance(input, object):
                output.append(eval('lambda T: {:f}'.format(input.to_base_units().magnitude)))
                outputStrs.append('lambda T: {:f}'.format(input.to_base_units().magnitude))
            else:
                raise ValueError('Unit cell property input has to be a single or '
                                 'cell array of numerics, function handles or strings which can be'
                                 'converted into a function handle!')

        return(output, outputStrs)

    @property
    def int_heat_capacity(self):
        """get int_heat_capacity

        Returns the anti-derrivative of the temperature-dependent heat
        :math:`c(T)` capacity function. If the int_heat_capacity
        property is not set, the symbolic integration is performed.

        """
        if hasattr(self, '_int_heat_capacity') and isinstance(self._int_heat_capacity, list):
            return self._int_heat_capacity
        else:
            self._int_heat_capacity = []
            self.int_heat_capacity_str = []
            try:
                T = Symbol('T')
                for i, hcs in enumerate(self.heat_capacity_str):
                    integral = integrate(hcs.split(':')[1], T)
                    self._int_heat_capacity.append(lambdify(T, integral))
                    self.int_heat_capacity_str.append('lambda T : ' + str(integral))

            except Exception as e:
                print('The sympy integration did not work. You can set the'
                      'analytical anti-derivative of the heat capacity'
                      'of your unit cells as lambda function of the temperature'
                      'T by typing UC.int_heat_capacity = lambda T: c(T)'
                      'where UC is the name of the unit cell object.')
                print(e)

        return self._int_heat_capacity

    @int_heat_capacity.setter
    def int_heat_capacity(self, int_heat_capacity):
        """set int_heat_capacity

        Set the integrated heat capacity manually when no sympy is
        installed.

        """
        self._int_heat_capacity, self.int_heat_capacity_str = self.check_cell_array_input(
                int_heat_capacity)

    @property
    def int_lin_therm_exp(self):
        """get int_lin_therm_exp

        Returns the anti-derrivative of the integrated
        temperature-dependent linear thermal expansion function. If the
        int_lin_therm_exp property is not set, the symbolic integration
        is performed.

        """

        if hasattr(self, '_int_lin_therm_exp') and isinstance(self._int_lin_therm_exp, list):
            return self._int_lin_therm_exp
        else:
            self._int_lin_therm_exp = []
            self.int_lin_therm_exp_str = []
            try:
                T = Symbol('T')
                for i, ltes in enumerate(self.lin_therm_exp_str):
                    integral = integrate(ltes.split(':')[1], T)
                    self._int_lin_therm_exp.append(lambdify(T, integral))
                    self.int_lin_therm_exp_str.append('lambda T : ' + str(integral))

            except Exception as e:
                print('The sympy integration did not work. You can set the'
                      'the analytical anti-derivative of the heat capacity'
                      'of your unit cells as lambda function of the temperature'
                      'T by typing UC.int_heat_capacity = lambda T: c(T)'
                      'where UC is the name of the unit cell object.')
                print(e)

        return self._int_lin_therm_exp

    @int_lin_therm_exp.setter
    def int_lin_therm_exp(self, int_lin_therm_exp):
        """set int_lin_therm_exp

        Set the integrated linear thermal expansion coefficient manually
        when no sympy installed.

        """
        self._int_lin_therm_exp, self.int_lin_therm_exp_str = self.check_cell_array_input(
                int_lin_therm_exp)

    def add_atom(self, atom, position):
        """ add_atom

        Adds an atomBase/atomMixed at a relative position of the unit
        cell.

        Update the mass, density and spring constant of the unit cell
        automatically:

        .. math:: \kappa = m \cdot (v_s / c)^2

        """
        position_str = ''
        # test the input type of the position
        if isfunction(position):
            raise ValueError('Please use string representation of function!')
            pass
        elif isinstance(position, str):
            try:
                position_str = position
                position = eval(position)
            except Exception as e:
                print('String input for unit cell property ' + position + ' \
                    cannot be converted to function handle!')
                print(e)
        elif isinstance(position, (int, float)):
            position_str = 'lambda strain: {:e}*(strain+1)'.format(position)
            position = eval(position_str)
        else:
            raise ValueError('Atom position input has to be a scalar, or string'
                             'which can be converted into a lambda function!')

        # add the atom at the end of the array
        self.atoms.append([atom, position, position_str])
        # increase the number of atoms
        self.num_atoms = self.num_atoms + 1

        self.mass = 0*u.kg
        for i in range(self.num_atoms):
            self.mass = self.mass + self.atoms[i][0].mass

        self.density = self.mass / self.volume
        # set mass per unit area (do not know if necessary)
        self.mass = self.mass * 1*u.angstrom**2 / self.area
        self.calc_spring_const()

    def add_multiple_atoms(self, atom, position, Nb):
        """add_multiple_atoms

        Adds multiple atomBase/atomMixed at a relative position of the
        unit cell.

        """
        for i in range(Nb):
            self.addAtom(atom, position)

    def calc_spring_const(self):
        """calc_spring_const

        Calculates the spring constant of the unit cell from the mass,
        sound velocity and c-axis

        .. math:: k = m \, \left(\\frac{v}{c}\\right)^2

        """
        self.spring_const[0] = (self._mass * (self._sound_vel/self._c_axis)**2)

    def get_acoustic_impedance(self):
        """get_acoustic_impedance"""
        Z = np.sqrt(self.spring_const[0] * self.mass/self.area**2)
        return Z

    def set_ho_spring_constants(self, HO):
        """set_ho_spring_constants

        Set the higher orders of the spring constant for anharmonic
        phonon simulations.

        """
        # reset old higher order spring constants
        self.spring_const = np.delete(self.spring_const, np.r_[1:len(self.spring_const)])
        self.spring_const = np.hstack((self.spring_const, HO))

    def get_atom_ids(self):
        """get_atom_ids

        Returns a cell array of all atom ids in the unit cell.

        """
        ids = []
        for i in range(self.num_atoms):
            if not self.atoms[i][0].id in ids:
                ids.append(self.atoms[i][0].id)

        return ids

    def get_atom_positions(self, *args):
        """get_atom_positions

        Returns a vector of all relative postion of the atoms in the
        unit cell.

        """
        if args:
            strain = args[0]
        else:
            strain = 0

        res = np.zeros([self.num_atoms])
        for i, atom in enumerate(self.atoms):
            res[i] = atom[1](strain)

        return res

    @property
    def a_axis(self):
        """float: in-plane a-axis [m]"""
        return Q_(self._a_axis, u.meter).to('angstrom')

    @a_axis.setter
    def a_axis(self, a_axis):
        """set.a_axis"""
        self._a_axis = a_axis.to_base_units().magnitude

    @property
    def b_axis(self):
        """float: in-plane b-axis [m]"""
        return Q_(self._b_axis, u.meter).to('angstrom')

    @b_axis.setter
    def b_axis(self, b_axis):
        """set.a_axis"""
        self._b_axis = b_axis.to_base_units().magnitude

    @property
    def c_axis(self):
        """float: out-of-plane c-axis [m]"""
        return Q_(self._c_axis, u.meter).to('angstrom')

    @c_axis.setter
    def c_axis(self, c_axis):
        """set.c_axis"""
        self._c_axis = c_axis.to_base_units().magnitude

    @property
    def mass(self):
        """float: mass of unit cell normalized to area of 1 Å² [kg]"""
        return Q_(self._mass, u.kg)

    @mass.setter
    def mass(self, mass):
        """set.mass"""
        self._mass = mass.to_base_units().magnitude

    @property
    def density(self):
        """float: density of the unitCell [kg/m³]"""
        return Q_(self._density, u.kg/u.m**3)

    @density.setter
    def density(self, density):
        """set.density"""
        self._density = density.to_base_units().magnitude

    @property
    def area(self):
        """
        float: area of epitaxial unit cells need for normation for
        correct intensities) [m²]

        """
        return Q_(self._area, u.m**2)

    @area.setter
    def area(self, area):
        """set.area"""
        self._area = area.to_base_units().magnitude

    @property
    def volume(self):
        """float: volume of unit cell [m³]"""
        return Q_(self._volume, u.m**3)

    @volume.setter
    def volume(self, volume):
        """set.volume"""
        self._volume = volume.to_base_units().magnitude

    @property
    def sound_vel(self):
        """float: sound velocity in the unit cell [m/s]"""
        return Q_(self._sound_vel, u.m/u.s)

    @sound_vel.setter
    def sound_vel(self, sound_vel):
        """set.sound_vel
        If the sound velocity is set, the spring constant is
        (re)calculated.
        """
        self._sound_vel = sound_vel.to_base_units().magnitude
        self.calc_spring_const()

    @property
    def phonon_damping(self):
        """float: damping constant of phonon propagation [kg/s]"""
        return Q_(self._phonon_damping, u.kg/u.s)

    @phonon_damping.setter
    def phonon_damping(self, phonon_damping):
        """set.phonon_damping"""
        self._phonon_damping = phonon_damping.to_base_units().magnitude

    @property
    def opt_pen_depth(self):
        """
        float: penetration depth for pump always for 1st subsystem
        light in the unit cell [m]

        """
        return Q_(self._opt_pen_depth, u.meter).to('nanometer')

    @opt_pen_depth.setter
    def opt_pen_depth(self, opt_pen_depth):
        """set.opt_pen_depth"""
        self._opt_pen_depth = opt_pen_depth.to_base_units().magnitude



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim.atoms import Atom, AtomMixed
import numpy as np
import pytest


def test_atom():
    Dy = Atom('Dy')
    assert Dy.symbol == 'Dy'
    assert Dy.id == 'Dy'
    assert Dy.ionicity == 0
    assert Dy.name == 'Dysprosium'
    assert Dy.atomic_number_z == 66
    assert Dy.mass_number_a == 162.5
    assert Dy.mass.magnitude == pytest.approx(2.698e-25)
    # check if python hash works the same on different systems
    assert np.array_equal(Dy.atomic_form_factor_coeff[10],
                          np.array([1.17404e+01, -9.99900e+03, 2.25052e-01]))
    assert np.array_equal(Dy.cromer_mann_coeff,
                          np.array([66.0, 0.0, 26.507, 17.6383, 14.5596, 2.96577, 2.1802, 0.202172,
                                    12.1899, 111.874, 4.29728]))

    Oxygen = Atom('O', id='myOxygen', ionicity=-1)
    assert Oxygen.symbol == 'O'
    assert Oxygen.id == 'myOxygen'
    assert Oxygen.ionicity == -1


def test_atom_mixed():
    DyTb = AtomMixed('DyTb')
    DyTb.add_atom(Atom('Dy', ID='Dy'), 0.4)
    DyTb.add_atom(Atom('Tb', ID='Tb'), 0.6)
    assert DyTb.name == 'DyTb'
    assert DyTb.id == 'DyTb'
    assert DyTb.atomic_number_z == 65.4
    assert DyTb.mass_number_a == 160.35518
    assert DyTb.mass.magnitude == pytest.approx(2.663e-25)



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='udkm1Dsim',
    version='0.1',
    packages=find_packages(),
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.dat', '*.nff'],
    },
    url='https://github.com/dschick/udkm1Dsim',  # Optional
    download_url='https://github.com/dschick/udkm1Dsim/archive/v0.1.tar.gz',
    install_requires=['numpy',
                      'pint',
                      'pytest',
                      'scipy',
                      'sympy',
                      'tabulate'],
    license='MIT',
    author='Daniel Schick, et. al.',
    author_email='schick.daniel@gmail.com',
    description='A Python Simulation Toolkit for ' \
                '1D Ultrafast Dynamics in Condensed Matter',  # Required
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of the udkm1Dsimpy module.
#
# udkm1Dsimpy is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2019 Daniel Schick

__all__ = ["Atom", "AtomMixed"]

__docformat__ = "restructuredtext"

import os
import numpy as np
import scipy.constants as constants
from tabulate import tabulate
from . import u


class Atom:
    """Atom

    The atom class is the smallest structural unit of which one can
    build larger structures. It holds real physical properties of atoms
    defined in the attrubutes section and can return parameters and data
    necessary for different simulation types.

    Args:
        symbol (str): symbol of the atom

    Keyword Args:
        id (str): id of the atom
        ionicity (int): ionicity of the atom

    Attributes:
        symbol (str): symbol of the element
        id (str): identifier of the atom, may differ from symbol and/or
           name
        name (str): name of the element (generic)
        atomic_number_z (int): Z atomic number
        mass_number_a (float): A atomic mass number
        ionicity (int): ionicity of the atom
        mass (float): mass of the atom [kg]
        atomic_form_factor_coeff (ndarray[float]): atomic form factor
           coefficients for energy-dependent atomic form factor
        cromer_mann_coeff (ndarray[float]): cromer-mann coefficients for
           angular-dependent atomic form factor

    References:

        .. [1] D. T. Cromer & J. B. Mann (1968). X-ray scattering
           factors computed from numerical Hartree–Fock wave functions.
           `Acta Crystallographica Section A, 24(2), 321–324.
           <http://www.doi.org/10.1107/S0567739468000550>`_
        .. [2] J. Als-Nielson, & D. McMorrow (2001).
           `Elements of Modern X-Ray Physics. New York: John Wiley &
           Sons, Ltd. <http://www.doi.org/10.1002/9781119998365>`_
        .. [3] B. L. Henke, E. M. Gullikson & J. C. Davis (1993).
           X-Ray Interactions: Photoabsorption, Scattering,
           Transmission, and Reflection at E = 50-30,000 eV, Z = 1-92.
           `Atomic Data and Nuclear Data Tables, 54(2), 181–342.
           <http://www.doi.org/10.1006/adnd.1993.1013>`_

    """

    def __init__(self, symbol, **kwargs):
        self.symbol = symbol
        self.id = kwargs.get('id', symbol)
        self.ionicity = kwargs.get('ionicity', 0)

        try:
            filename = os.path.join(os.path.dirname(__file__),
                                    'parameters/elements/elements.dat')
            symbols = np.genfromtxt(filename, dtype='U2', usecols=(0))
            elements = np.genfromtxt(filename, dtype='U15, i8, f8', usecols=(1, 2, 3))
            [rowidx] = np.where(symbols == self.symbol)
            element = elements[rowidx[0]]
        except Exception as e:
            print('Cannot load element specific data from elements data file!')
            print(e)

        self.name = element[0]
        self.atomic_number_z = element[1]
        self.mass_number_a = element[2]
        self.mass = self.mass_number_a*constants.atomic_mass*u.kg
        self.atomic_form_factor_coeff = self.read_atomic_form_factor_coeff()
        self.cromer_mann_coeff = self.read_cromer_mann_coeff()

    def __str__(self):
        """String representation of this class"""
        output = {'parameter': ['id', 'symbol', 'name', 'atomic number Z', 'mass number A', 'mass',
                                'ionicity', 'Cromer Mann coeff', '', ''],
                  'value': [self.id, self.symbol, self.name, self.atomic_number_z,
                            self.mass_number_a, '{:.4~P}'.format(self.mass), self.ionicity,
                            np.array_str(self.cromer_mann_coeff[0:4]),
                            np.array_str(self.cromer_mann_coeff[4:8]),
                            np.array_str(self.cromer_mann_coeff[8:])]}

        return 'Atom with the following properties\n' + \
               tabulate(output, colalign=('right',), tablefmt="rst", floatfmt=('.2f', '.2f'))

    def read_atomic_form_factor_coeff(self):
        """read_atomic_form_factor_coeff

        The atomic form factor :math:`f` in dependence from the energy
        :math:`E` is read from a parameter file given by [3]_.

        """
        filename = os.path.join(os.path.dirname(__file__),
                                'parameters/atomicFormFactors/{:s}.nff'.format(
                                        self.symbol.lower()))
        try:
            f = np.genfromtxt(filename, skip_header=1)
        except Exception as e:
            print('File {:s} not found!\nMake sure the path /parameters/atomicFormFactors/ is in'
                  'your search path!',
                  filename)
            print(e)

        return f

    @u.wraps(None, (None, 'eV'), strict=True)
    def get_atomic_form_factor(self, E):
        """get_atomic_form_factor

        Returns the complex atomic form factor

        .. math:: f(E)=f_1-i f_2

        for the energy :math:`E` [eV].

        Convention of Ref. [2]_ (p. 11, footnote) is a negative :math:`f_2`

        """
        # interpolate the real and imaginary part in dependence of E
        f1 = np.interp(E, self.atomic_form_factor_coeff[:, 0], self.atomic_form_factor_coeff[:, 1])
        f2 = np.interp(E, self.atomic_form_factor_coeff[:, 0], self.atomic_form_factor_coeff[:, 2])

        return f1 - f2*1j

    def read_cromer_mann_coeff(self):
        """read_cromer_mann_coeff

        The Cromer-Mann coefficients (Ref. [1]_) are read from a
        parameter file and are returned in the following order:

        .. math:: a_1\; a_2\; a_3\; a_4\; b_1\; b_2\; b_3\; b_4\; c

        """
        filename = os.path.join(os.path.dirname(__file__),
                                'parameters/atomicFormFactors/cromermann.txt')
        try:
            cm = np.genfromtxt(filename, skip_header=1,
                               usecols=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11))
        except Exception as e:
            print('File {:s} not found!\nMake sure the path'
                  '/parameters/atomicFormFactors/ is in your search path!',
                  filename)
            print(e)

        return cm[(cm[:, 0] == self.atomic_number_z) & (cm[:, 1] == self.ionicity)][0]

    @u.wraps(None, (None, 'eV', 'angstrom**-1'), strict=True)
    def get_cm_atomic_form_factor(self, E, qz):
        """get_cm_atomic_form_factor

        Returns the atomic form factor :math:`f` in dependence of the
        energy :math:`E` [J] and the :math:`z`-component of the
        scattering vector :math:`q_z` [Å :math:`^{-1}`] (Ref. [1]_).
        Since the CM coefficients are fitted for :math:`q_z` in
        [Å :math:`^{-1}`] we have to convert it before!

        See Ref. [2]_ (p. 235).

        .. math:: f(q_z,E) = f_{CM}(q_z) + \delta f_1(E) -i f_2(E)

        :math:`f_{CM}(q_z)` is given in Ref. 1:

        .. math::

            f_{CM}(q_z) = \sum(a_i \, \exp(-b_i \, (q_z/4\pi)^2))+ c

        :math:`\delta f_1(E)` is the dispersion correction:

        .. math::

            \delta f_1(E) = f_1(E) - \left(\sum^4_i(a_i) + c\\right)

        Thus:

        .. math:: f(q_z,E) = \sum(a_i \, \exp(b_i \, q_z/2\pi))
           + c + f_1(E)-i f_2(E) - \left(\sum(a_i) + c\\right)

        .. math:: f(q_z,E) = \sum(a_i \, \exp(b_i \, q_z/2\pi))
           + f_1(E) -i f_2(E) - \sum(a_i)

        """

        f_cm = np.dot(self.cromer_mann_coeff[0:3],
                      np.exp(np.dot(-self.cromer_mann_coeff[4:7],
                                    (qz/(4*np.pi))**2))) + self.cromer_mann_coeff[8]

        return f_cm + self.get_atomic_form_factor(E*u.eV) -\
            (np.sum(self.cromer_mann_coeff[0:3]) + self.cromer_mann_coeff[8])


class AtomMixed(Atom):
    """AtomMixed

    The AtomMixed class is sub class of Atom and enables mixed atoms for
    certain alloys and stochiometric mixtures. All properties of the
    included sub-atoms of class atomBase are averaged and weighted with
    their stochiometric ratio

    Args:
        symbol (str): symbol of the atom

    Keyword Args:
        id (str): id of the atom
        ionicity (int): ionicity of the atom

    Attributes:
        symbol (str): symbol of the element
        id (str): identifier of the atom, may differ from symbol and/or
           name
        name (str): name of the element (generic)
        atomic_number_z (int): Z atomic number
        mass_number_a (float): A atomic mass number
        ionicity (int): ionicity of the atom
        mass (float): mass of the atom [kg]
        atomic_form_factor_coeff (ndarray[float]): atomic form factor
           coefficients for energy-dependent atomic form factor
        cromer_mann_coeff (ndarray[float]): cromer-mann coefficients for
           angular-dependent atomic form factor

    """

    def __init__(self, symbol, **kwargs):
        self.symbol = symbol
        self.id = kwargs.get('id', symbol)
        self.name = kwargs.get('name', symbol)
        self.ionicity = 0
        self.atomic_number_z = 0
        self.mass_number_a = 0
        self.mass = 0
        self.atoms = []
        self.num_atoms = 0
        self.cromer_mann_coeff = np.array([])

    def __str__(self):
        """String representation of this class"""
        output = []
        for i in range(self.num_atoms):
            output.append([self.atoms[i][0].name, '{:.1f} %'.format(self.atoms[i][1]*100)])

        return (super().__str__()
                + '\n{:d} Constituents:\n'.format(self.num_atoms)
                + tabulate(output, colalign=('right',), floatfmt=('.2f', '.2f')))

    def add_atom(self, atom, fraction):
        """add_atom

        Add an Atom instance with its stochiometric fraction to the AtomMixed instance.
        """
        self.atoms.append([atom, fraction])
        self.num_atoms = self.num_atoms + 1
        # calculate the mixed atomic properties of the atomMixed instance
        self.atomic_number_z = self.atomic_number_z + fraction * atom.atomic_number_z
        self.mass_number_a = self.mass_number_a + fraction * atom.mass_number_a
        self.mass = self.mass + fraction * atom.mass
        self.ionicity = self.ionicity + fraction * atom.ionicity

    def get_atomic_form_factor(self, E):
        """get_atomic_form_factor

        Returns the mixed energy dependent atomic form factor.
        """
        f = 0
        for i in range(self.num_atoms):
            f += self.atoms[i][0].get_atomic_form_factor(E) * self.atoms[i][1]

        return f

    def get_cm_atomic_form_factor(self, E, qz):
        """get_cm_atomic_form_factor

        Returns the mixed energy and angle dependent atomic form factor.
        """
        f = 0
        for i in range(self.num_atoms):
            f += self.atoms[i][0].get_cm_atomic_form_factor(E, qz) * self.atoms[i][1]

        return f
